@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -101,8 +101,10 @@ void perl_run_blocks(I32 oldscope, AV *list)
 	CV *cv = (CV*)*av_fetch(list, i, FALSE);
 	SV* atsv = ERRSV;
 
+	MARK_WHERE("END block", (SV*)cv);
 	PUSHMARK(stack_sp);
 	perl_call_sv((SV*)cv, G_EVAL|G_DISCARD);
+	UNMARK_WHERE;
 	(void)SvPV(atsv, len);
 	if (len) {
 	    if (list == beginav)
@@ -127,14 +129,14 @@ void mod_perl_clear_rgy_endav(request_rec *r, SV *sv)
 	SV *entry = *hv_fetch(mod_perl_endhv, key, klen, FALSE);
 	AV *av;
 	if(!SvTRUE(entry) && !SvROK(entry)) {
-	    MP_TRACE(fprintf(stderr, "endav is empty for %s\n", r->uri));
+	    MP_TRACE_g(fprintf(stderr, "endav is empty for %s\n", r->uri));
 	    return;
 	}
 	av = (AV*)SvRV(entry);
 	av_clear(av);
 	SvREFCNT_dec((SV*)av);
 	(void)hv_delete(mod_perl_endhv, key, klen, G_DISCARD);
-	MP_TRACE(fprintf(stderr, 
+	MP_TRACE_g(fprintf(stderr, 
 			 "clearing END blocks for package `%s' (uri=%s)\n",
 			 key, r->uri)); 
     }
@@ -149,7 +151,7 @@ void perl_run_rgy_endav(char *s)
     dTHR;
 
     if(!rgystash) {
-	MP_TRACE(fprintf(stderr, 
+	MP_TRACE_g(fprintf(stderr, 
         "Apache::Registry::curstash not set, can't run END blocks for %s\n",
 			 s));
 	return;
@@ -181,8 +183,8 @@ void perl_run_rgy_endav(char *s)
 	}
     }
 
-    MP_TRACE(fprintf(stderr, 
-	     "running %d END blocks for %s\n", rgyendav ? AvFILL(rgyendav)+1 : 0, s));
+    MP_TRACE_g(fprintf(stderr, 
+	     "running %d END blocks for %s\n", rgyendav ? (int)AvFILL(rgyendav)+1 : 0, s));
     if((endav = rgyendav)) 
 	perl_run_blocks(scopestack_ix, endav);
     if(rgyendav)
@@ -194,11 +196,14 @@ void perl_run_rgy_endav(char *s)
 void perl_run_endav(char *s)
 {
     dTHR;
+    I32 n = 0;
+    if(endav)
+	n = AvFILL(endav)+1;
+
+    MP_TRACE_g(fprintf(stderr, "running %d END blocks for %s\n", 
+		       (int)n, s));
     if(endav) {
-	save_hptr(&curstash);
 	curstash = defstash;
-	MP_TRACE(fprintf(stderr, "running %d END blocks for %s\n", 
-			 AvFILL(endav)+1, s));
 	call_list(scopestack_ix, endav);
     }
 }
@@ -254,32 +259,58 @@ void perl_reload_inc(void)
     (void)hv_iterinit(orig_inc);
     while((val = hv_iternextsv(orig_inc, &key, &klen))) {
 	perl_require_pv(key);
-	MP_TRACE(fprintf(stderr, "reloading %s\n", key));
+	MP_TRACE_g(fprintf(stderr, "reloading %s\n", key));
     }
 
     LEAVE;
 }
 
+I32 perl_module_is_loaded(char *name)
+{
+    I32 retval = FALSE;
+    SV *key = perl_module2file(name);
+    if((key && hv_exists_ent(GvHV(incgv), key, FALSE)))
+	retval = TRUE;
+    if(key)
+	SvREFCNT_dec(key);
+    return retval;
+}
+
+SV *perl_module2file(char *name)
+{
+    SV *sv = newSVpv(name,0);
+    char *s;
+    for (s = SvPVX(sv); *s; s++) {
+	if (*s == ':' && s[1] == ':') {
+	    *s = '/';
+	    Move(s+2, s+1, strlen(s+2)+1, char);
+	    --SvCUR(sv);
+	}
+    }
+    sv_catpvn(sv, ".pm", 3);
+    return sv;
+}
+
 int perl_require_module(char *mod, server_rec *s)
 {
     dTHR;
     SV *sv = sv_newmortal();
     sv_setpvn(sv, "require ", 8);
-    MP_TRACE(fprintf(stderr, "loading perl module '%s'...", mod)); 
+    MP_TRACE_d(fprintf(stderr, "loading perl module '%s'...", mod)); 
     sv_catpv(sv, mod);
     perl_eval_sv(sv, G_DISCARD);
     if(s) {
 	if(perl_eval_ok(s) != OK) {
-	    MP_TRACE(fprintf(stderr, "not ok\n"));
+	    MP_TRACE_d(fprintf(stderr, "not ok\n"));
 	    return -1;
 	}
     }
     else if(SvTRUE(ERRSV)) {
-	MP_TRACE(fprintf(stderr, "not ok\n"));
+	MP_TRACE_d(fprintf(stderr, "not ok\n"));
 	return -1;
     }
 
-    MP_TRACE(fprintf(stderr, "ok\n"));
+    MP_TRACE_d(fprintf(stderr, "ok\n"));
     return 0;
 }
 
@@ -299,11 +330,11 @@ int perl_load_startup_script(server_rec *s, pool *p, char *script, I32 my_warn)
     I32 old_warn = dowarn;
 
     if(!script) {
-	MP_TRACE(fprintf(stderr, "no PerlScript to load\n"));
+	MP_TRACE_d(fprintf(stderr, "no Perl script to load\n"));
 	return OK;
     }
 
-    MP_TRACE(fprintf(stderr, "attempting to load `%s'\n", script));
+    MP_TRACE_d(fprintf(stderr, "attempting to require `%s'\n", script));
     dowarn = my_warn;
     curstash = defstash;
     perl_do_file(script);
@@ -311,6 +342,23 @@ int perl_load_startup_script(server_rec *s, pool *p, char *script, I32 my_warn)
     return perl_eval_ok(s);
 } 
 
+array_header *perl_cgi_env_init(request_rec *r)
+{
+    table *envtab = r->subprocess_env; 
+    char *tz = NULL; 
+
+    add_common_vars(r); 
+    add_cgi_vars(r); 
+
+    if ((tz = getenv("TZ")) != NULL)
+	table_set(envtab, "TZ", tz);
+
+    table_set(envtab, "PATH", DEFAULT_PATH);
+    table_set(envtab, "GATEWAY_INTERFACE", PERL_GATEWAY_INTERFACE);
+
+    return table_elts(envtab);
+}
+
 void perl_clear_env(void)
 {
     char *key; 
@@ -341,7 +389,7 @@ void mod_perl_init_ids(void)  /* $$, $>, $), etc */
     euid = (int)geteuid(); 
     gid  = (int)getgid(); 
     egid = (int)getegid(); 
-    MP_TRACE(fprintf(stderr, 
+    MP_TRACE_g(fprintf(stderr, 
 		     "perl_init_ids: uid=%d, euid=%d, gid=%d, egid=%d\n",
 		     uid, euid, gid, egid));
 #endif
@@ -352,7 +400,7 @@ int perl_eval_ok(server_rec *s)
     dTHR;
     SV *sv = ERRSV;
     if(SvTRUE(sv)) {
-	MP_TRACE(fprintf(stderr, "perl_eval error: %s\n", SvPV(sv,na)));
+	MP_TRACE_g(fprintf(stderr, "perl_eval error: %s\n", SvPV(sv,na)));
 	mod_perl_error(s, SvPV(sv, na));
 	return -1;
     }
@@ -363,7 +411,7 @@ int perl_sv_is_http_code(SV *errsv, int *status)
 {
     int i=0, http_code=0, retval = FALSE;
     char *errpv;
-    char cpcode[3];
+    char cpcode[4];
 
     if(!SvTRUE(errsv) || !ERRSV_CAN_BE_HTTP)
 	return FALSE;
@@ -383,7 +431,7 @@ int perl_sv_is_http_code(SV *errsv, int *status)
      * if they're not all digits, $@ is not an HTTP code
      */
     if(http_code != 3) {
-	MP_TRACE(fprintf(stderr, 
+	MP_TRACE_g(fprintf(stderr, 
 			 "mod_perl: $@ doesn't look like an HTTP code `%s'\n", 
 			 errpv));
 	return FALSE;
@@ -393,10 +441,11 @@ int perl_sv_is_http_code(SV *errsv, int *status)
     if(SvCUR(errsv) == http_code)
 	return TRUE;
 
-    MP_TRACE(fprintf(stderr, "mod_perl: possible $@ HTTP code `%s'\n", errpv));
+    ap_cpystrn((char *)cpcode, errpv, 4);
 
-    /* XXX, use new ap_strncpy */
-    strncpy((char *)cpcode, errpv, 3);
+    MP_TRACE_g(fprintf(stderr, 
+		     "mod_perl: possible $@ HTTP code `%s' (cp=`%s')\n", 
+		     errpv,cpcode));
 
     if((SvCUR(errsv) == 4) && (*(SvEND(errsv) - 1) == '\n')) {
 	/* nothin but 3 digit code and \n */
@@ -405,6 +454,7 @@ int perl_sv_is_http_code(SV *errsv, int *status)
     else {
 	char *tmp = errpv;
 	tmp += 3;
+#ifndef PERL_MARK_WHERE
 	if(strNE(SvPVX(GvSV(curcop->cop_filegv)), "-e")) {
 	    SV *fake = newSV(0);
 	    sv_setpv(fake, ""); /* avoid -w warning */
@@ -416,14 +466,15 @@ int perl_sv_is_http_code(SV *errsv, int *status)
 
 	    SvREFCNT_dec(fake);
 	}
-	else if(strnEQ(tmp, " at ", 4) && instr(errpv, " line "))
+#endif
+	if(!retval && strnEQ(tmp, " at ", 4) && instr(errpv, " line "))
 	    /* well, close enough */
 	    retval = TRUE;
     }
 
     if(retval == TRUE) {
     	*status = atoi(cpcode);
-	MP_TRACE(fprintf(stderr, 
+	MP_TRACE_g(fprintf(stderr, 
 			 "mod_perl: $@ is an HTTP code `%d'\n", *status));
     }
 
@@ -455,3 +506,77 @@ void perl_incpush(char *p)
 	av_push(GvAV(incgv), libdir);
     }
 }
+
+#ifdef PERL_MARK_WHERE
+/* XXX find the right place for this! */
+static SV *perl_sv_name(SV *svp)
+{
+    SV *sv = Nullsv;
+    SV *RETVAL = Nullsv;
+
+    if(svp && SvROK(svp) && (sv = SvRV(svp))) {
+	switch(SvTYPE(sv)) {
+	case SVt_PVCV:
+	    RETVAL = newSV(0);
+	    gv_fullname(RETVAL, CvGV(sv));
+	    break;
+
+	default:
+	    break;
+	}
+    }
+    else if(svp && SvPOK(svp)) {
+	RETVAL = newSVsv(svp);
+    }
+
+    return RETVAL;
+}
+
+void mod_perl_mark_where(char *where, SV *sub)
+{
+    SV *name = Nullsv;
+    if(curcop->cop_line) {
+#if 0
+	fprintf(stderr, "already know where: %s line %d\n",
+		SvPV(GvSV(curcop->cop_filegv),na), curcop->cop_line);
+#endif
+	return;
+    }
+
+    SAVESPTR(curcop->cop_filegv);
+    SAVEI16(curcop->cop_line);
+
+    if(sub) 
+	name = perl_sv_name(sub);
+
+    sv_setpv(GvSV(curcop->cop_filegv), "");
+    sv_catpvf(GvSV(curcop->cop_filegv), "%s subroutine `%_'", where, name);
+    curcop->cop_line = 1;
+
+    if(name)
+	SvREFCNT_dec(name);
+}
+#endif
+
+#if MODULE_MAGIC_NUMBER < 19971226
+char *ap_cpystrn(char *dst, const char *src, size_t dst_size)
+{
+
+    char *d, *end;
+
+    if (!dst_size)
+        return (dst);
+
+    d = dst;
+    end = dst + dst_size - 1;
+
+    while ((d < end) && (*d++ = *src++))
+        ;	/* nop, the while does it all */
+
+    *d = '\0';	/* always null terminate */
+
+    return (d);
+}
+
+#endif
+
